"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_rs.src.losses.pairwise_hinge_loss import (
    PairwiseHingeLoss as PairwiseHingeLoss,
)
from keras_rs.src.losses.pairwise_logistic_loss import (
    PairwiseLogisticLoss as PairwiseLogisticLoss,
)
from keras_rs.src.losses.pairwise_mean_squared_error import (
    PairwiseMeanSquaredError as PairwiseMeanSquaredError,
)
from keras_rs.src.losses.pairwise_soft_zero_one_loss import (
    PairwiseSoftZeroOneLoss as PairwiseSoftZeroOneLoss,
)
